''Copyright (c) 2010, HMCsoft and Ebben Feagan
''
''All rights reserved.
''
''Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
''
''    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
''    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
''    * Neither the name of HMCsoft nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
''
''THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
''"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
''LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
''A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
''CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
''EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
''PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
''PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
''LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
''NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
''SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'Compiles LOC and a host of other information of a directory tree for later processing.

#define fbext_NoBuiltinInstanciations() 1

'ext includes
#include once "ext/misc.bi"
#include once "ext/xml.bi"
#include once "ext/strings.bi"
#include once "ext/debug.bi"
#include once "ext/file/file.bi"
#include once "ext/options.bi"

'fb includes
#include once "file.bi"

dim shared as string base_href
dim shared as integer ln
dim shared as ext.xml.node ptr last_was_in
dim shared as ext.File ff
    dim num_code as uinteger
    dim num_comment as uinteger
    dim num_ws as uinteger

	enum linetype
		unknown
		whitespace
		comment
		code
		type_
		include
		macro
	end enum

	enum statuscodes
		default = 0
		inFunction
		multiLineDeclare
		inType
	end enum

declare sub AnalyzeFile( byval xnode as ext.xml.node ptr )
declare function readNextLine( ) as string
declare sub RecurseInto( byval xnode as ext.xml.node ptr )
declare function getLinetype( byval xnode as ext.xml.node ptr, byref lstr as const string ) as linetype
declare function isFunction( byref lstr as const string ) as string
declare function getFunctionSignature( byref lstr as const string ) as string
declare function findFunction( byval xnode as ext.xml.node ptr, byref lstr as const string ) as ext.xml.node ptr
declare function isType( byref lstr as const string ) as string
declare function isPreProcessor( byval xnode as ext.xml.node ptr, byref lstr as const string ) as linetype

using ext
using misc
using strings
using xml


''::: MAIN :::
dim as xml.tree xtree
dim as options.Parser opt

''::: OPTIONS :::
opt.setHelpHeader( !"FBEXT - KLOC\n" & _
			"FreeBASIC Source Code Analyzer." )
var oo = 	opt.addOption("o","output-file",true,true,false,,"Sets the file to output XML data to." )
var oi = 	opt.addOption("i","input-file",true,true,false,,"Analyze a single file." )
var od = 	opt.addOption("d","directory",true,true,false,,"Analyze the contents of a directory (includes subdirectories)." )
var lang = 	opt.addOption("l","language",true,true,false,,"Sets the souce language, currently supported is: FreeBASIC." )
opt.setHelpFooter( !"The i and d options are mutually exclusive, you may only choose one or the other.\n" & _
	!"Default behaviour is outputs all data to STDOUT and runs with the current directory as the target.\n" & _
	"Copyright (c) 2010, The FreeBASIC Extended Library Development Group" )

opt.parse(__FB_ARGC__,__FB_ARGV__)

if opt.isSet(oi) = true AND opt.isSet(od) = true then
	opt.showHelp
	end 1
end if

var outputf = ""
if opt.isSet(oo) then outputf = opt.getArg(oo)
if opt.isSet(od) then
	base_href = opt.getArg(od)
else
	base_href = curdir()
end if

'FBEXT_DPRINT("Counting lines in: " & curdir())

xtree.root->appendChild("ext-kloc")
xtree.root->child("ext-kloc")->attribute("base-href") = base_href
base_href = curdir
xtree.root->child("ext-kloc")->attribute("run-date") = date
xtree.root->child("ext-kloc")->attribute("start-time") = time
xtree.root->child("ext-kloc")->appendChild("directory")->attribute("name")= "/"

if not(opt.isSet(oi)) then
	RecurseInto( xtree.root->child("ext-kloc")->child("directory") )
else
	var fzz = xtree.root->child("ext-kloc")->child("directory")->appendChild("file")
	fzz->attribute("name") = opt.getArg(oi)
	AnalyzeFile( fzz )
end if

chdir base_href

if outputf = "" then
	open cons for output as #1
else
	open outputf for binary ACCESS WRITE as #1
end if

print #1, xtree

close

''::: END MAIN :::

sub RecurseInto( byval xnode as ext.xml.node ptr )


    dim as string this_dir

	this_dir = xnode->attribute("name")

	replace( this_dir, "//", "/" )

	if right(this_dir,1) = "/" then
		this_dir = left(this_dir,len(this_dir)-1)
	end if

	if len(this_dir) > 0 then
	chdir this_dir
	end if

	'FBEXT_DPRINT("In directory: " & curdir())


    for n as FILE_ITER = FILE_ITER("*",&h10) to ""

	    var temp = n.filename()
		'FBEXT_DPRINT(":: " & temp)
        if (len(temp) > 0) ANDALSO (temp[0] <> asc(".")) then

            xnode->appendChild("directory")->attribute("name") = temp

        end if
    next n


        for m as FILE_ITER = FILE_ITER("*.bas",&h21) to ""

            var temp = m.filename()

            if (len(temp) > 0) ANDALSO (temp[0] <> asc(".")) then

                 var new_file = xnode->appendChild("file")
                 new_file->attribute("name") = temp

                AnalyzeFile( new_file )

            end if

        next m

        for m as FILE_ITER = FILE_ITER("*.bi",&h21) to ""

            var temp = m.filename()

            if (len(temp) > 0) ANDALSO (temp[0] <> asc(".")) then


                var new_file = xnode->appendChild("file")
                new_file->attribute("name") = temp

                AnalyzeFile( new_file )


            end if

        next m

	for r as integer = 0 to xnode->children("directory")-1
		RecurseInto( xnode->child("directory",r) )
	next

	if base_href <> curdir() then
	chdir ".."
	end if

end sub

function ReadNextLine( ) as string

	if not(ff.eof) then
		ln += 1
		return ff.linput
	else
		return ""
	end if

end function

sub AnalyzeFile( byval xnode as ext.xml.node ptr )

    dim file_con as string
    num_code = 0
    num_comment = 0
    num_ws = 0
    dim file_size as ulongint = 0
    dim file_name as string
    dim as uinteger max_line = 0
    dim as uinteger avg_line = 0
    dim as integer file_enc = 0
    dim as double file_time = 0.0
    dim last_func as ext.xml.node ptr

    file_con = ""



	dim status as statuscodes = default

	var this_dir = xnode->attribute("name")

	file_name = this_dir
	file_time = filedatetime( file_name )

	ff = ext.File(file_name,ext.File.R)

    if ff.open = false then
		'FBEXT_DPRINT("Processing file: " & file_name)
		file_enc = fileattr( ff.handle, fbFileAttrEncoding )
        file_size = ff.lof
        ln = 1
        while not ff.eof
            file_con = ReadNextLine
            if len(trim(file_con)) > 0 then
				var tlen = len(file_con)
				if tlen > max_line then max_line = tlen
				avg_line += tlen
			end if

        wend

	else

		xnode->attribute("error") = "Unable to open file"
		ff.close
		return

	end if
    ff.close

	xnode->attribute("timestamp") = str(file_time)

	var strenc = ""
	select case file_enc
	case fbFileEncodASCII
		strenc = "ASCII"
	case fbFileEncodUTF16
		strenc = "UTF16"
	case fbFileEncodUTF32
		strenc = "UTF32"
	case else
		strenc = "UTF8"
	end select

	xnode->attribute("encoding") = strenc

	xnode->attribute("filesize") = str(file_size)

    var tloc = xnode->appendChild( "loc" )
    tloc->attribute("comment") = str(num_comment)
    tloc->attribute("code") = str(num_code)
    tloc->attribute("whitespace") = str(num_ws)
    tloc->attribute("total") = str(num_ws + num_code + num_comment)

    xnode->appendChild( "linewidth" )
    xnode->child("linewidth")->attribute("average") = str( avg_line \ (num_code + num_comment + num_ws) )
    xnode->child("linewidth")->attribute("max") = str( max_line )

end sub

function getLinetype( byval xnode as ext.xml.node ptr ) as linetype

	var lstr = trim(ReadNextLine)

	if lstr = "" then return linetype.whitespace

	if left(lstr,1) = "'" or left(lcase(lstr),3) = "rem" then return linetype.comment

	var detect = isPreProcessor( xnode, lstr )
	if detect >= linetype.include then return linetype.include

	detect = isType( xnode, lstr )
	if detect = linetype.type_ then return linetype.type_

	detect = isfunction( xnode, lstr )

	return detect

	/'
						'proc defines first
						if (detect[1] = asc("[") orelse detect[1] = asc("{") _
							orelse detect[1] = asc("]") orelse detect[1] = asc("}")) _
							AND len(detect)>3 then

							select case detect[0]
							case asc(":") 'subdecl
								var nt = xnode->appendChild("sub")
								last_func = nt
								nt->attribute("name") = right(detect,len(detect)-3)
								nt->attribute("defined-on") = str(ln)
								select case detect[1] 'calling convention
								case asc("[") 'fbcall
									nt->attribute("call-conv") = "fbcall"
								case asc("{") 'cdecl
									nt->attribute("call-conv") = "cdecl"
								case asc("]") 'pascal
									nt->attribute("call-conv") = "pascal"
								case asc("}") 'stdcall
									nt->attribute("call-conv") = "stdcall"
								end select
								if detect[2] = asc("<") then
									nt->attribute("exported") = "true"
								else
									nt->attribute("exported") = "false"
								end if
							case asc(";") 'funcdecl
								var nt = xnode->appendChild("function")
								last_func = nt
								nt->attribute("name") = right(detect,len(detect)-3)
								nt->attribute("defined-on") = str(ln)
								select case detect[1] 'calling convention
								case asc("[") 'fbcall
									nt->attribute("call-conv") = "fbcall"
								case asc("{") 'cdecl
									nt->attribute("call-conv") = "cdecl"
								case asc("]") 'pascal
									nt->attribute("call-conv") = "pascal"
								case asc("}") 'stdcall
									nt->attribute("call-conv") = "stdcall"
								end select
								if detect[2] = asc("<") then
									nt->attribute("exported") = "true"
								else
									nt->attribute("exported") = "false"
								end if
							end select
							status = inFunction

						else 'then declares
							select case detect[0]
							case asc(":") 'subdecl
								var nt = xnode->appendChild("sub")
								nt->attribute("name") = right(detect,len(detect)-1)
								nt->attribute("declared-on") = str(ln)
							case asc(";") 'funcdecl
								var nt = xnode->appendChild("function")
								nt->attribute("name") = right(detect,len(detect)-1)
								nt->attribute("declared-on") = str(ln)
							end select
						end if
					end if
					end if
					if (left(trim(lcase(file_con)),7) = "end sub" ORELSE _
							left(trim(lcase(file_con)),7) = "end fun") AND _
							status = inFunction then
						if last_func <> 0 then last_func->attribute("length") = str(ln-valint(last_func->attribute("defined-on"))-1)
						last_func = 0
						status = default
					end if
                    num_code += 1
                end if
            else
                num_ws += 1
            end if
	'/
end function

private function findFunction( byval xnode as ext.xml.node ptr, byref lstr as const string ) as ext.xml.node ptr

		var asloc = instrrev(lcase(lstr),"as")
		var outstr = ""
		var l = asloc
		getword:
		for n as integer = asloc+2 to len(lstr)
			if lstr[n] <> 32 then
				outstr = outstr & chr(lstr[n])
			else
				exit for
			end if
			l = n
		next
		if lcase(outstr) = "const" then
			asloc = l -1
			outstr &= " "
			goto getword
		end if
		if l < len(lstr)-1 then
			if instr(l,lcase(lstr),"ptr") > 0 then
				outstr &= "*"
			end if
		end if

		return outstr

end function

function getFunctionSignature( byref lstr as const string ) as string

	dim as ext.bool isSub = true
	if instr(lcase(lstr),"sub") <1 then isSub = false

	var lparens = instr(lstr,"(")

	''detect sub subname, and func funcname as return type
	if lparens < 1 then
		if isSub then return "( )"
		return ""'getReturnType(lstr) & " ( )"
	end if




end function

	enum typeofparam
		declare_
		name_
		callconv
		as_
		ppc
		variadic
		proctype
		visibility
	end enum

private function typeOfParameter( byref x as const string ) as typeofparam

	select case lcase(x)
	case "declare"
		return typeofparam.declare_
	case "sub", "function", "operator", "constructor", "property", "destructor"
		return typeofparam.proctype
	case "as"
		return typeofparam.as_
	case "byval", "byref", "bydesc"
		return typeofparam.ppc
	case "cdecl", "stdcall", "pascal"
		return typeofparam.callconv
	case "..."
		return typeofparam.variadic
	case "public", "private", "export"
		return typeofparam.visibility
	case else
		return typeofparam.name_
	end select

end function

function isFunction( byval xnode as ext.xml.node ptr, byref lstr as const string ) as linetype



	dim as ubyte lasttok = 0
	var curtok = ext.strings.tokenize( lstr, !" \t_,()", @lasttok )

	select case lcase(curtok)
	case "declare", "sub", "function", "public", "private", "operator", "property", "constructor", "destructor"
	case else
		return linetype.code
	end select

	while (curtok <> "")
		select case lcase(curtok)
		case "declare"

		case "sub", "function", "public", "private", "operator", "property", "constructor", "destructor"

		end select
		curtok = ext.strings.tokenize( , !" \t_,()", @lasttok )
	wend

	if instr(lcase(lstr),"declare") > 0 then
		'proc declare
		var newlstr = trim(right(lstr,len(lstr)-7))
		var iss = instr(lcase(newlstr),"sub")
		var isf = instr(lcase(newlstr),"function")
		if iss > 0 then
			newlstr = trim(mid(newlstr,iss+3))
			var wh = instr(newlstr,"(")
			var wh2 = instr(iss,newlstr," ")
			if wh2 < wh then swap wh, wh2
			if wh > 0 then
				newlstr = mid( newlstr, 1, wh-1 )
			end if
			return ":" & newlstr
		elseif isf > 0 then
			newlstr = trim(mid(newlstr,isf+8))
			var wh = instr(newlstr,"(")
			var wh2 = instr(isf, newlstr," ")
			if wh2 < wh then swap wh, wh2
			if wh > 0 then
				newlstr = mid( newlstr, 1, wh-1 )
			end if
			return ";" & newlstr
		else
			return ""
		end if

	else

		if (instr(lcase(lstr)," sub ") > 0 orelse instr(lcase(lstr)," function ") > 0) _
		orelse (instr(lcase(lstr),"sub ") > 0 orelse instr(lcase(lstr),"function ") > 0) _
		ANDALSO left(lcase(lstr),4) <> "type" then
			'its probably a proc
			if (left(lcase(lstr),6) <> "public" orelse left(lcase(lstr),7) <> "private") AND _
			(instr(lcase(lstr),"sub")>1 ORELSE instr(lcase(lstr),"function")>1) then return ""
			var ret = ""
			var cal = "[" 'FBCALL 0[, CDECL 1{, PASCAL 2], STDCALL 3}
			var exported = "," ', - 1<
			var whz = 5
			var wh = instr(lstr," sub ")
			if wh <1 then wh = instr(lstr,"sub ")
			if wh <1 then
				wh = instr(lstr," function ")
				if wh <1 then wh = instr(lstr,"function ")
				whz = 10
				ret = ";"
			else
				ret = ":"
			end if
			var newlstr = trim(mid(lstr,wh))
			var wh3 = instr(wh,newlstr,"(")
			var wh2 = instr(wh,newlstr," ")
			if wh3 < wh2 then swap wh2, wh3
			wh3 = instr(wh2+1,newlstr,"(")
			var wh4 = instr(wh2+1,newlstr," ")
			if wh4 < wh3 then swap wh3, wh4
			var fname = mid(newlstr,wh2+1,wh3-whz)

			if instr(lcase(lstr),"cdecl")>0 then cal = "{"
			if instr(lcase(lstr),"pascal")>0 then cal = "]"
			if instr(lcase(lstr),"stdcall")>0 then cal = "}"
			if instr(lcase(lstr),"export")>0 then exported = "<"

			var wlstr = lcase(lstr)
			if instr(wlstr,"end sub") > 0 ORELSE _
				instr(wlstr,"end function") > 0 ORELSE _
				instr(wlstr,!"\"function\"")> 0 ORELSE _
				instr(wlstr,!"\"sub\"")>0 then return ""

			return ret & cal & exported & fname
		end if

	end if

	return ""

end function

function isType( byval xnode as ext.xml.node ptr, byref lstr as const string ) as linetype

	if left(lcase(lstr),4) = "type" then

		var ret = trim(right(lstr,len(lstr)-4))
		var spa = instr(ret," ")
		if spa > 0 then
			ret = mid( ret, 1, spa-1 )
		end if

		var nt = xnode->appendChild("type")
			nt->attribute("name") = ret
			nt->attribute("defined-on") = str(ln)

		var aaa = instr(lcase(lstr),"as")
		if aaa<1 then
			var wh2 = instr(lstr," ")
			if wh2 <0 then len(lstr)
			ret = mid(lstr, aaa+3, (wh2-1)-(aaa+3))
			nt->attribute("redirects-to") = ret
		else
			'TODO: parse members
		end if

		return linetype.type_

	end if

	return linetype.unknown

end function

function isPreProcessor( byval xnode as ext.xml.node ptr, byref lstr as const string ) as linetype

	if lstr[0] = asc("#") then

		if left(lcase(lstr),8) = "#include" ORELSE left(lcase(lstr),9) = "# include" then

			var wh = instr(lstr, !"\"" )
			var wh2 = instr(wh+1, lstr, !"\"" )
			var ret = mid(lstr, wh+1, (wh2-1)-wh)
			var nt = xnode->appendChild("include",node_type_e.text)
							nt->setText = ret
							nt->attribute("line") = str(ln)
							if instr(lstr, "once")>0 and instr(lstr,"once")<wh then _
										nt->attribute("pragma") = "once"

			return linetype.include

		end if

		if left(lcase(lstr),7) = "#inclib" ORELSE left(lcase(lstr),8) = "# inclib" then

			var wh = instr(lstr, !"\"" )
			var wh2 = instr(wh+1, lstr, !"\"" )
			var ret = mid(lstr, wh+1, (wh2-1)-wh)

			var nt = xnode->appendChild("library",node_type_e.text)
				nt->setText = ret
				nt->attribute("line") = str(ln)

			return linetype.include

		end if
	end if

	return linetype.unknown

end function
